# -----------------------------------------------------------------------------
# From the book "Managing Enterprise Active Directory Services"
# ISBN: 0-672-32125-4
# Copyright (C) 2002 by Addison-Wesley
# Script by Robbie Allen
# 
# You have a royalty-free right to use, modify, reproduce and distribute
# this script (and/or any modified version) in any way you find useful,
# provided that you agree that Addison-Wesley or Robbie Allen has no
# warranty, obligations or liability for the script. If you modify
# the script, you must retain this copyright notice.
# -----------------------------------------------------------------------------
$|++; # auto-flush output

use strict;
use Net::LDAP;
use Net::LDAP::LDIF;
use Win32::NetResource;
use Win32::TieRegistry(Delimiter=>"/");

my $server  = 'dc1';
my $user    = 'administrator@xyz.com';
my $user_nbt = 'XYZ\administrator';
my $passwd  = 'password';

my $ldif_file = $ARGV[0] || 'app1.ldif';

my $ldap = Net::LDAP->new($server); 
my $bind = $ldap->bind($user, password => $passwd);
die "Could not bind to server: ", $bind->error if $bind->code;

####################################
# Get the schemaNamingContext value
####################################
my $rootdse = $ldap->search(base   => '',
                            filter => '(objectclass=*)',
                            scope  => 'base',
                            attrs => ['schemaNamingContext']);
my $schema_dn = ($rootdse->entries)[0]->get_value('schemaNamingContext');
print "Schema Container: $schema_dn\n";

####################################
# Get the Schema fsmo
####################################
# Get the DN of the NTDS Settings DN for the FSMO holder
my $schema = $ldap->search(base   => $schema_dn,
                           filter => '(&(objectclass=dMD)(fsmoRoleOwner=*))',
                           scope  => 'subtree',
                           attrs => ['fsmoRoleOwner']);
my $fsmo_dn = ($schema->entries)[0]->get_value('fsmoRoleOwner');

# Strip off the RDN of the NTDS Settings DN
$fsmo_dn =~ s/^cn=NTDS Settings,//i;

# Get the dNSHostname attribute 
my $fsmo = $ldap->search(base   => $fsmo_dn,
                         filter => '(objectclass=*)',
                         scope  => 'base',
                         attrs => ['dNSHostName']);
my $schemafsmo = ($fsmo->entries)[0]->get_value('dNSHostName');
print "Schema FSMO: $schemafsmo\n";

####################################
# Bind against the Schema FSMO
####################################
unless ($server eq $schemafsmo) {
  $ldap->unbind;
  $ldap = Net::LDAP->new($schemafsmo); 
  $ldap->bind($user, password => $passwd);
}

####################################
# Verify schema extensions do not exist
####################################
my $schema = $ldap->schema();
my $ldif = Net::LDAP::LDIF->new( $ldif_file, "r", onerror => 'undef' );

print "Verifying LDIF entries\n";
my $ldf_err;
while( not $ldif->eof() ) {
    my $entry = $ldif->read_entry();

    if ( $ldif->error() ) {
       print "\nError msg: ",$ldif->error(),"\n";
       print "\nError lines:\n",$ldif->error_lines(),"\n";
       $ldf_err++;
       last;
    }
    # Only want to check the entries that are to be added
    elsif ($entry->exists( 'cn' ) and $entry->changetype eq 'add') {
       my $cn = $entry->get_value('cn');
       if ($schema->is_objectclass( $cn )) {
          print "Existing class: $cn\n";
          $ldf_err++;
       }
       elsif ($schema->is_attribute( $cn )) {
          print "Existing attribute: $cn\n";
          $ldf_err++;
       }
    }
}
$ldif->done();
die "Aborting, error with LDIF file\n" if $ldf_err;

####################################
# Update registry of Schema FSMO
####################################
my $reg_key  = "System/CurrentControlSet/Services/NTDS/";
   $reg_key .= "Parameters/Schema Update Allowed";
my $set_value  = '0x00000001';
$user = 'XYZ\administrator';
my %conn_info = (Type => RESOURCETYPE_ANY, 
                 RemoteName => "\\\\$schemafsmo\\IPC\$");
Win32::NetResource::AddConnection(\%conn_info, $passwd, $user, 0)
      or die "Could not add connection: $^E\n";
my $regObj = $Registry->{"//$schemafsmo/LMachine"} 
          or die "Could not connect: $^E";
print "Enabling schema updates\n";
$regObj->{$reg_key} = [ $set_value, 'REG_DWORD' ] 
               or die "Could not modify registry: $^E\n";

####################################
# Import LDIF file
####################################
$ldif = Net::LDAP::LDIF->new( $ldif_file, "r", onerror => 'undef' );
print "Inserting LDIF entries";
while( not $ldif->eof() ) {
    my $entry = $ldif->read_entry();
    if ( $ldif->error() ) {
       print "\nError msg: ",$ldif->error(),"\n";
       print "\nError lines:\n",$ldif->error_lines(),"\n";
       last;
    }
    else {
       print ".";
       my $res = $entry->update( $ldap );
       if ($res->code) {
          print "Error inserting entry: ", $res->error,"\n";
          last;
       }
    }
}
print "done\n";
$ldif->done();

####################################
# Update registry of Schema FSMO
####################################
my $unset_value  = '0x00000000';
print "Disabling schema updates\n";
$regObj->{$reg_key} = [ $unset_value, 'REG_DWORD' ]
           or die "Could not modify registry: $^E\n";
Win32::NetResource::CancelConnection("\\\\$schemafsmo", 0, 0)
           or die "Could not cancel IPC connection: $^E\n";
